
#include <stdio.h>
#include <stdlib.h>

#include "base.h"
#include "glifos.h"
#include "grises.h"
#include "grises0.h"
#include "imag.h"
#include "imagio.h"
#include "imagx.h"
#include "inclina.h"
#include "inclina0.h"
// #include "inclfont.h"
#include "opcns.h"

#define NSEGMENS 20000

char * Mode = "ocre.segmenta2";

typedef struct temarco{
  int dx, dy;
  int fx, fy;
  int tx, ty;
  } * tmarco;

void liberaMarcosDeSegmento (tmarco * marcos, int nm)
{
  int n;
  for (n=1; n<=nm; n++)
    free (marcos[n]) ;
}

int ** matalloc (int tx, int ty)
{
  int ** m;
  int y;
  m = (int **) malloc (ty * sizeof(int *));
  for (y=0; y<ty; y++)
    m[y] = (int *) malloc (tx * sizeof(int));
  return m ;
}

void matfree (int ** m, int ty)
{
  int y;
  for (y=0; y<ty; y++)
    free (m[y]);
  free (m) ;
}

void escribeFondo (int ** fondo, int tx, int ty)
{
  int x, y;
  for (y=0; y<ty+2; y++) {
    for (x=0; x<tx+2; x++)
      if (fondo[y][x] == 0)
        printf (" .");
      else
        printf ("%2d", fondo[y][x] % 100);
    printf ("\n");
    }
}

void escribeFondoYmarco (int ** fondo, int tx, int ty, tmarco marco, int incl)
{
  int x, y;
  int delx;
  int dx, dy, fx, fy;
  dx = marco->dx + 1;
  dy = marco->dy + 1;
  fx = marco->fx + 1;
  fy = marco->fy + 1;
  delx = inclf_desvx(incl, fy-dy);
  printf ("%4d %4d\n", incl, delx);
  for (y=0; y<ty+2; y++) {
    for (x=0; x<tx+2; x++)
      if (((x==dx)      && (y==dy)) ||
          ((x==dx+delx) && (y==fy)) ||
          ((x==fx)      && (y==dy)) ||
          ((x==fx+delx) && (y==fy)))
        printf (" +");
      else if (fondo[y][x] == 0)
        printf (" .");
      else
        printf ("%2d", fondo[y][x]);
    printf ("\n");
    }
}

// --------------------  --------------------

void ponEquiv (int *equivalencias, int nseg, int n1, int n2)
{
  int n, en1, en2;
  en1 = min(equivalencias[n1],equivalencias[n2]);
  en2 = max(equivalencias[n1],equivalencias[n2]);
  if (en1 != en2)
    for (n=1; n<=nseg; n++)
      if (equivalencias[n] == en2)
        equivalencias[n] = en1;
}

void  comprimeEquivalencias (int * equivalencias, int *pnseg)
{
  int n, ns, nseg;
  ns = 0;
  nseg = *pnseg ;
  for (n=1; n <= nseg; n++){
    if (equivalencias[n] == n) {
      equivalencias[n] = ++ ns;
      }
    }
  *pnseg = ns;
}

void renumeraFondo (int **fondo, int *equivalencias, int tx, int ty)
{
  int xf, yf;
  for (yf=1; yf<(ty+1); yf++)
  for (xf=1; xf<(tx+1); xf++)
    fondo[yf][xf] = equivalencias[fondo[yf][xf]];
}

//  Compara cada pixel con los  situados al norte y al oeste,
//                                       al noroeste y al noreste.
//  Sigue enlaces mas delgados y produce menos segmentos
void marcaFondo (int **fondo, int * pnseg, tvent vent, int umbral)
{
  char **pix;
  int tx, ty;
  int x, y, xf, yf;
  int nseg;
  int equivalencias[NSEGMENS];
  if (d_valori >= 12) {
    printf ("ocre.segmenta2.marcaFondo:\n");
    describeVentana (vent);
    }
  pix = vent->pix ;
  tx  = vent->tx ;
  ty  = vent->ty ;
  // printf ("tx:%4d ty:%4d\n", tx, ty);
  nseg = 0;
  // ponemos a cero los bordes superior,izquierdo y derecho de  fondo
  for (x=0; x<(tx+2); x++)
    fondo[0][x] = 0;
  for (y=0; y<(ty+1); y++)
    fondo[y][0] = 0;
  for (y=0; y<(ty+1); y++)
    fondo[y][tx+1] = 0;
  // ponemos a cero el borde inferior de  fondo  ??
  for (x=0; x<(tx+2); x++)
    fondo[ty+1][x] = 0;
  // recorremos toda la imagen ..
  for (y=0; y<ty; y++) {
    yf = y + 1;
    for (x=0; x<tx; x++) {
      xf = x + 1;
      if ((pix[y][x] & 0xFF) > umbral) 
        fondo[yf][xf] = 0;
      else {
        int etiq;
        etiq = 0;
        // oeste
        if (fondo[yf][xf-1] != 0) {
          etiq = fondo[yf][xf-1] ;
          }
        // noroeste
        if (fondo[yf-1][xf-1] != 0) {
          if (etiq != 0)
            ponEquiv (equivalencias, nseg, etiq, fondo[yf-1][xf-1]) ;
          etiq = fondo[yf-1][xf-1] ;
          }
        // norte
        if (fondo[yf-1][xf] != 0) {
          if (etiq != 0)
            ponEquiv (equivalencias, nseg, etiq, fondo[yf-1][xf]) ;
          etiq = fondo[yf-1][xf] ;
          }
        // noreste
        if (fondo[yf-1][xf+1] != 0) {
          if (etiq != 0)
            ponEquiv (equivalencias, nseg, etiq, fondo[yf-1][xf+1]) ;
          etiq = fondo[yf-1][xf+1] ;
          }
        if (etiq != 0)
          fondo[yf][xf] = etiq;
        else  /* aislado */ {
          fondo [yf][xf] = ++ nseg ;
          equivalencias[nseg] = nseg;
          if (nseg > NSEGMENS) {
              printf ("ocre.segmenta2.marcaFondo: error, nseg = %4d\n",nseg);
              exit (2);
            }
          }
        }
      }
    } 
  // cada segmento : a un solo valor
  // printf ("ocre.segmenta2.marcaFondo: sigue 1\n"); // XXx1
  equivalencias[0] = 0;
  renumeraFondo (fondo, equivalencias, tx, ty);  // ?
  // los segmentos : consecutivos
  comprimeEquivalencias (equivalencias, &nseg);
  renumeraFondo (fondo, equivalencias, tx, ty);
  //
  * pnseg = nseg ;
}

// --------------------  --------------------

// dx .. fy  a partir de 0
void limitesSegmentoRec (int *dx, int *dy, int *fx, int *fy,
                         int ** fondo, int n, tvent vent)
{
  int tx, ty, xf, yf;
  int encontrado;
  tx = vent->tx;
  ty = vent->ty;
  /* norte */
  encontrado = FALSE;
  for (yf=1; yf<(ty+1) && ! encontrado; yf++)
  for (xf=1; xf<(tx+1) && ! encontrado; xf++)
    if (fondo[yf][xf] == n){
      encontrado = TRUE;
      *dy = yf - 1;
      }
  /* sur */ 
  encontrado = FALSE;
  for (yf=ty; yf>0 && ! encontrado; yf--)
  for (xf=tx; xf>0 && ! encontrado; xf--)
    if (fondo[yf][xf] == n){
      encontrado = TRUE;
      *fy = yf - 1;
      }
  /* oeste */
  encontrado = FALSE;
  for (xf=1; xf<(tx+1) && ! encontrado; xf++)
  for (yf=1; yf<(ty+1) && ! encontrado; yf++)
    if (fondo[yf][xf] == n){
      encontrado = TRUE;
      *dx = xf - 1;
      }
  /* este */
  encontrado = FALSE;
  for (xf=tx; xf>0 && ! encontrado; xf--)
  for (yf=ty; yf>0 && ! encontrado; yf--)
    if (fondo[yf][xf] == n){
      encontrado = TRUE;
      *fx = xf - 1;
      }
}

// dx .. fy  a partir de 0
Boolean limitesSegmentoInc (int *dx, int *dy, int *fx, int *fy,
                         int ** fondo, int n, tvent vent)
{
  char * funce = "limitesSegmentoInc";
  int tx, ty, xf, yf;
  int encontrado;
  int xi, incl, xp, xu;
  int x1, x2, y1, y2;
  int xpy, xuy;
  tx = vent->tx;
  ty = vent->ty;
//   escribeFondo (fondo, tx, ty);
  if (d_valori >= 20) {
    printf ("%s.%s: 0 n:%4d tx:%4d ty:%4d\n", Mode, funce, n, tx, ty);
    }
  //
  incl = vent->incl;
  pon_vprim_vult (incl, ty, tx, &xp, &xu) ;
  //  printf ("%s.%s: 1 xp:%4d xu:%4d incl:%4d\n", Mode, funce, xp, xu, incl);
  // norte
  encontrado = FALSE;
  for (yf=1; yf<(ty+1) && ! encontrado; yf++){
    xpy = xp + inclf_desvx(incl,yf);
    xuy = xu + inclf_desvx(incl,yf);
    for (xf=xpy+1; xf<(xuy+1) && ! encontrado; xf++)
      if (fondo[yf][xf] == n){
        encontrado = TRUE;
        y1 = yf;
        }
      }
  if (! encontrado){
    // printf ("%s.%s: error, no encotrado %3d norte \n", Mode, funce, n);   // aviso/warning ?
    return FALSE;
    }
  // sur
  encontrado = FALSE;
  for (yf=ty; yf>0 && ! encontrado; yf--){
    xpy = xp + inclf_desvx(incl,yf);
    xuy = xu + inclf_desvx(incl,yf);
    for (xf=xpy+1; xf<(xuy+1) && ! encontrado; xf++)
      if (fondo[yf][xf] == n){
        encontrado = TRUE;
        y2 = yf;
        }
      }
  if (! encontrado){
    // printf ("%s.%s: error, no encotrado %3d sur \n", Mode, funce, n);   // aviso/warning ?
    return FALSE;
    }
  // oeste
  encontrado = FALSE;
  for (xf=xp+1; xf<=(xu+1) && ! encontrado; xf++)
  for (yf=y1; yf<=y2 && ! encontrado; yf++) {
    xi = xf + inclf_desvx(incl, yf);
    if (fondo[yf][xi] == n){
      encontrado = TRUE;
      x1 = xf + inclf_desvx(incl,y1) ;
      }
    }
  if (! encontrado){
    // printf ("%s.%s: error, no encotrado %3d oeste \n", Mode, funce, n);   // aviso/warning ?
    return FALSE;
    }
  // este
  encontrado = FALSE;
  for (xf=xu+1; xf>=(xp+1) && ! encontrado; xf--)
  for (yf=y1; yf<=y2 && ! encontrado; yf++) {
    xi = xf + inclf_desvx(incl, yf);
    if (fondo[yf][xi] == n){
      encontrado = TRUE;
      x2 = xf + inclf_desvx(incl,y1) ; 
      }
    }
  if (! encontrado){
    // printf ("%s.%s: error, no encotrado %3d este \n", Mode, funce, n);   // aviso/warning ?
    return FALSE;
    }
  //
  *dx = x1 - 1 ;
  *fx = x2 - 1 ;
  *dy = y1 - 1 ;
  *fy = y2 - 1 ;
  if (d_valori >= 10) {
    printf ("segmenta2.limitesSegmentoInc %4d %4d %4d %4d\n",
                                          x1-1, y1-1, x2-1, y2-1);
    }
  return TRUE;
}

void limitesSegmento (int *dx, int *dy, int *fx, int *fy,
                      int ** fondo, int n, tvent vent)
{
  Boolean encontrados;
  if (vent->incl == 0)
    limitesSegmentoRec (dx, dy, fx, fy, fondo, n, vent);
  else  /* vent->incl != 0 */ {
    encontrados = limitesSegmentoInc (dx, dy, fx, fy, fondo, n, vent);
    if (! encontrados)
      limitesSegmentoRec (dx, dy, fx, fy, fondo, n, vent);
    }
}
  
// --------------------  --------------------

// funcion no usada  !?!?
// para ayudar, borrar luego
/*
void quitaOtrosSegmentos (tvent ventl, tvent vent, int **fondo, int n,
                          int dxs, int dys, int fxs, int fys, int umbral2)
{
  int x, y, xl, yl, xf, yf;
  char ** pix, ** pixl;
  pix  = vent->pix;
  pixl = ventl->pix;
  for (y=dys; y<=fys; y++) {
    yf = y+1;
    yl = y-dys+1;
    for (x=dxs; x<=fxs; x++) {
      xf = x+1;
      xl = x-dxs+1;
      if ((fondo[yf][xf] == 0) || (fondo[yf][xf] == n))
        pixl [yl][xl] = pix [y][x];
      else
        pixl [yl][xl] = umbral2 ;
      }
    }
  // pon bordes 
  for (xl=0; xl<ventl->tx; xl++) {
    pixl [0]          [xl] = umbral2;
    pixl [ventl->ty-1][xl] = umbral2;
    }
  for (yl=0; yl<ventl->ty; yl++) {
    pixl [yl][0]           = umbral2;
    pixl [yl][ventl->tx-1] = umbral2;
    }
}
*/

int K2 = 0;

//   dados
// vent      ventana de partida
// fondo     sombras con numero 
// enGrupo   grupo de una sombra
// n         grupo seleccionado
// marcoG    marco del grupo
// umbral2   para rellenar
//   obtiene
// ventl     ventana limpia quitando sombras de otro grupo
void quitaOtrosSegmentosGrupo (tvent ventl, tvent vent, int **fondo, int * enGrupo,
                               int n, tmarco marcoG, int umbral2)
{
  int dx, dy, fx, fy;    // desplazamineto y final del marco
  int x, y;              // en vent
  int xl, yl;            // en marco y en ventl
  int xf, yf;            // en fondo
  int incl;
  // int desvxit;           // desviacion x por inclinacion 
  int desvxity;          // desviacion x por inclinacion
  // int xlpr1, xlur1;      // x (limp) primero y ultimo, restado_1 (en el fondo)
  // char ** pix;
  char ** pixl;
  K2++;
  // printf ("quitaOtrosSegmentosGrupo --- K2: %4d\n", K2);
  dx = marcoG->dx;
  dy = marcoG->dy;
  fx = marcoG->fx;
  fy = marcoG->fy;
  // pix  = vent->pix;
  pixl = ventl->pix;
  incl = vent->incl;
  // printf("---2-\n");
  // describeVentana (ventl);                                          // 
  // desvxit = inclf_absDesvx(incl, fy - dy);
  // printf ("quitaOtrosSegmentosGrupo dx:%4d dy:%4d fx:%4d fy:%4d incl:%5d desvxit:%4d \n", dx, dy, fx, fy, incl, desvxit); 
   // printf ("quitaOtrosSegmentosGrupo dx:%4d dy:%4d fx:%4d fy:%4d incl:%5d\n", dx, dy, fx, fy, incl); //
   // escribeFondoYmarco (fondo, vent->tx, vent->ty, marcoG, incl);  // 
  for (y=dy; y<=fy; y++) {
    yf = y+1;
    yl = y-dy+1;
    desvxity = inclf_desvx(incl, y-dy);
    for (x=dx+desvxity; x<=fx+desvxity; x++) {
      xf = x+1 ; 
      xl = x-dx+1 ;   //  ??  mar abr  3 11:39:27 CEST 2012
      // xl = x+1 ;      //  ??  mar abr  3 11:39:27 CEST 2012
    // pon_vprim_vult(incl, fy-dy+1, fx-dx+1, &xlpr1, &xlur1);
    // printf ("y:%3d yf:%3d yl:%3d  xlpr1:%3d xlur1:%3d\n",
    //          y,    yf,    yl,     xlpr1,    xlur1      );
    // for (xl=xlpr1+1; x<=(xlur1+1); xl++) {
    //   xf = xl + dx;
      if ((fondo[yf][xf] != 0) && (enGrupo[fondo[yf][xf]] != n)) {
        // sombra de otro grupo
        pixl [yl][xl]   = umbral2 ;
        if ((yl+1) < yf)
          pixl [yl+1][xl] = umbral2 ;
        if ((yl-1) >= 0)
          pixl [yl-1][xl] = umbral2 ;
        if ((xl+1) < xf)
          pixl [yl][xl+1] = umbral2 ;
        if ((xl-1) >= 0)
          pixl [yl][xl-1] = umbral2 ;
        }
      }
    }
  // printf("---3-\n");
  // describeVentana (ventl);                                          // 
  // pon bordes
  for (xl=0; xl<ventl->tx; xl++) {
    pixl [0]          [xl] = umbral2;
    pixl [ventl->ty-1][xl] = umbral2;
    }
  for (yl=0; yl<ventl->ty; yl++) {
    pixl [yl][0]           = umbral2;
    pixl [yl][ventl->tx-1] = umbral2;
    }
}

void ponUmbrales (int *umbral, int *umbral2, tvent vent, int porcNeg)
{
  tdcaracHistoGr carHisGr;
  carHisGr = carhisgrDeVentana (vent);
  *umbral  = (carHisGr->xnegro  * porcNeg +
              carHisGr->xblanco * (100 - porcNeg) ) / 100;
  *umbral2 = carHisGr->xblanco ;
}

// --------------------  --------------------

tmarco marcoDelGrupo (int * enGrupo, tmarco * marcos, int nm, int n, int incl)
{
  tmarco marcoG;
  int m;
  int dx=0, dy=0, fx=0, fy=0;  // '= 0' para evitar avisos(warnings)
  int primero;
  marcoG = (tmarco) malloc (sizeof(struct temarco));
  primero = TRUE;
  for (m=1; m<=nm; m++) 
    if (enGrupo[m] == n) {
      if (primero) {
        dx = marcos[m]->dx ;
        dy = marcos[m]->dy ;
        fx = marcos[m]->fx ;
        fy = marcos[m]->fy ;
        primero = FALSE ;
        }
      else {
        if (incl == 0) {
          dy = min(dy, marcos[m]->dy) ;
          fy = max(fy, marcos[m]->fy) ;
          dx = min(dx, marcos[m]->dx) ;
          fx = max(fx, marcos[m]->fx) ;
          }
        else  /* incl != 0 */ {
          int dxn1, dxnn, dy2;
          fy = max(fy, marcos[m]->fy) ;              // sur nuevo
          dy2 = min(dy, marcos[m]->dy) ;             // norte nuevo
          dxn1 = inclf_desvx(-incl, marcos[m]->dy - dy2); // delx marco[m]
          dxnn = inclf_desvx(-incl,            dy - dy2); // delx ant.
          dy = dy2 ;
          dx = min(dx+dxnn, marcos[m]->dx + dxn1);
          fx = max(fx+dxnn, marcos[m]->fx + dxn1);
          // printf ("ocre.segmenta2.marcoDelGrupo.enfor %4d %4d %4d %4d %5d %5d\n", marcos[m]->dx, marcos[m]->dy, marcos[m]->fx, marcos[m]->fy, dxn1, dxnn);
          // printf ("ocre.segmenta2.marcoDelGrupo.enfor %4d %4d %4d %4d %5d\n", dx, dy, fx, fy, incl);
          }
        }
      if (d_valori >= 30) {
        printf ("ocre.segmenta2.marcoDelGrupo.enfor %4d %4d %4d %4d %5d\n",
                                                    dx, dy, fx, fy, incl);
        }
      }
  if (incl != 0) {
    int dxs;
    dxs = inclf_desvx(incl, fy-dy);
    dx = max(-dxs, dx);
    }
  marcoG->dx = dx ;
  marcoG->dy = dy ;
  marcoG->fx = fx ;
  marcoG->fy = fy ;
  if (d_valori >= 10) {
    printf ("ocre.segmenta2.marcoDelGrupo %4d %4d %4d %4d %5d\n",
                                          dx, dy, fx, fy, incl);
    }
  return marcoG ;
}

// esta funcion no se usa !?
int unicosSegmentosI (int ** fondo, int * enGrupo, int n, tmarco marco, int incl)
{
  int xf, yf;
  int delx;
  int dx, dy, fx, fy;
  int encontrado;
  dx = marco->dx ;
  dy = marco->dy ;
  fx = marco->fx ;
  fy = marco->fy ;
  encontrado = FALSE;
  for (yf=dy+1; yf<(fy+1) && ! encontrado; yf++) {
    delx = inclf_desvx(incl, yf-dy-1);
    for (xf=dx+delx+1; xf<(fx+delx+1) && ! encontrado; xf++)
      if ((enGrupo[fondo[yf][xf]] != n) &&
          (        fondo[yf][xf]  != 0)) {
        encontrado = TRUE;
        }
    }
  return !encontrado;
}

int unicosSegmentos (int ** fondo, int * enGrupo, int n, int dx, int dy, int fx, int fy)
{
  int xf, yf;
  int encontrado;
  encontrado = FALSE;
  for (yf=dy+1; yf<(fy+1) && ! encontrado; yf++)
  for (xf=dx+1; xf<(fx+1) && ! encontrado; xf++)
    if ((enGrupo[fondo[yf][xf]] != n) &&
        (        fondo[yf][xf]  != 0)) {
      encontrado = TRUE;
      }
  return !encontrado;
}

tvent ventanaGrupoSegmentos (tvent vent, int **fondo, int * enGrupo,
                                  tmarco * marcos, int nm, int n, int umbral2)
{
  int dx, dy, tx, ty;    // los tamanyos en la ventana (si hay inclinacion, mayores que en el marco)
  tmarco marcoG;
  tvent vents, ventsAmpl, ventLimp;   // ventana segmento, id. ampliada, id. limpia
  timagf nimag;
  // printf ("ocre.segmenta2.ventanaGrupoSegmentos: entrando\n");     //
  // describeVentana (vent);                                          // 
  marcoG = marcoDelGrupo (enGrupo, marcos, nm, n, vent->incl);
  dx = marcoG->dx ;
  dy = marcoG->dy ;
  tx = marcoG->fx - dx + 1;
  ty = marcoG->fy - dy + 1;
  // printf ("=1 dx=%3d dy=%3d tx=%3d ty=%3d\n", dx, dy, tx, ty);      // 
  if (vent->incl != 0) {
    int adxs;
    adxs = inclf_absDesvx(vent->incl, ty-1);
    if (vent->incl < 0)
      dx -= adxs ;
    tx += adxs ;
  }
  // printf ("=2 dx=%3d dy=%3d tx=%3d ty=%3d\n", dx, dy, tx, ty);      // 
  ventanaEnVentana (&vents, vent, dx, dy, tx, ty) ;
  // printf ("ocre.segmenta2.ventanaGrupoSegmentos:  venEnVen1\n");   // 
  // describeVentana (vents);                                         // 
  vents->incl = vent->incl ;
  // if ( !unicosSegmentosI (fondo, enGrupo, n, marcoG, vent->incl) ) {
  if ( !unicosSegmentos (fondo, enGrupo, n, dx, dy, marcoG->fx, marcoG->fy) ) {
    // ventanaDeVentana (&ventsAmpl, vent, dx-1, dy-1, tx+2, ty+2) ;
    ventanaDeVentanaCabeMen (&ventsAmpl, vent, dx-1, dy-1, tx+2, ty+2) ;
             // un pixel de margen alrededor de 'vents' para poder comparar
    copiaImagenDeVentana ( &nimag, ventsAmpl);
    // describeVentana (nimag->ven0);                                          // 
    quitaOtrosSegmentosGrupo (nimag->ven0, vent, fondo, enGrupo, n, marcoG, umbral2);
    // describeVentana (nimag->ven0);                                          // 
    // printf ("ocre.segmenta2.ventanaGrupoSegmentos:  venEnVen2\n");   // 
    // ventanaEnVentana (&ventLimp , nimag->ven0, 1, 1, tx, ty);
    ventanaEnVentanaCabeMen (&ventLimp , nimag->ven0, 1, 1, tx, ty);
    // printf ("ocre.segmenta2.ventanaGrupoSegmentos:  venEnVen3\n");   // 
    ventLimp->incl = vent->incl ;
    ponerCarhisgrenVentana (ventLimp, carhisgrDeVentana(vent));
    vents->venlimpia = ventLimp ;
    }
  free (marcoG);
  return vents;
}

int marcosDeSegmento (tmarco * marcos, tvent vent, int ** fondo, int nseg)
{
  int dxs, dys, txs, tys, fxs, fys, tam;
  int n, ns;
  int nuevoNum [NSEGMENS];
  nuevoNum [0] = 0;
  ns = 0;
  for (n=1; n<=nseg; n++) {
    limitesSegmento (&dxs, &dys, &fxs, &fys, fondo, n, vent);
    txs = fxs - dxs + 1;
    tys = fys - dys + 1;
    tam = txs * tys ;
    if (tam > 2) {
      /* se ignoran los segmentos pequenyos, 2 es arbitrario */
      /* pon segmento en marcos */
      ns ++;
      if (ns >= NSEGMENS) {
        printf ("segmenta2.marcosDeSegmento: error, demasiados marcos\n");
        exit (1);
        }
      marcos [ns] = (tmarco) malloc (sizeof(struct temarco)) ;
      marcos [ns]->dx = dxs;
      marcos [ns]->dy = dys;
      marcos [ns]->fx = fxs;
      marcos [ns]->fy = fys;
      marcos [ns]->tx = txs;
      marcos [ns]->ty = tys;
      nuevoNum [n] = ns;
      if (d_valori >= 20) {
        printf ("segmenta2.marcosDeSegmento %3d %4d %4d %4d %4d %5d\n",
                                            ns, dxs, dys, fxs, fys, tam);
        }
      }
    else
      nuevoNum [n] = 0;
    }
  renumeraFondo (fondo, nuevoNum, vent->tx, vent->ty);
  return ns ;
}

// --------------------  --------------------

int enVertical (tmarco * marcos, int m1, int m2, int incl)
{
  int cdg1x ;
  int cdg1x_enm2;
  cdg1x = (marcos[m1]->dx + marcos[m1]->fx) / 2 ;
  if (marcos[m2]->dy > marcos[m1]->dy)
    cdg1x_enm2 = inclf_desvx(incl, marcos[m2]->dy - marcos[m1]->dy) + cdg1x;
  else
    cdg1x_enm2 = inclf_desvx(-incl, marcos[m1]->dy - marcos[m2]->dy) + cdg1x;
//   printf ("segmenta2.enVertical: %2d %2d, %4d %4d %4d\n", m1, m2, marcos[m2]->dx, cdg1x_enm2, marcos[m2]->fx) ; //
  return (cdg1x_enm2 >= marcos[m2]->dx) && (cdg1x_enm2 <= marcos[m2]->fx) ;
}

int disjuntosMar (tmarco * marcos, int m1, int m2)
{
  int disjuntosH, disjuntosV ;
  disjuntosH = max(marcos[m1]->dx,marcos[m2]->dx) >
               min(marcos[m1]->fx,marcos[m2]->fx) ;  // refinar
  disjuntosV = max(marcos[m1]->dy,marcos[m2]->dy) >
               min(marcos[m1]->fy,marcos[m2]->fy) ;
  return disjuntosH || disjuntosV ;
}

int agrupaMarcos (int * enGrupo, tmarco * marcos, int nm, int base, int alturax, int incl)
{
  int ng;
  int m, m2;
  // asignaGrupo a los marcos que esten entre la base y la alturax
  //               es decir, que no esten solo encima o solo abajo
  ng = 0 ;
  for (m=1; m<=nm; m++)
    // if ((marcos[m]->dy < base) && (marcos[m]->fy > alturax))
    if ((marcos[m]->dy < base) && (marcos[m]->fy > alturax) &&
        ((marcos[m]->fy-marcos[m]->dy+1) > ((base-alturax+1)/3)) )
      enGrupo [m] = ++ ng ;
    else
      enGrupo [m] = 0;
  // agrupa, si puede, los marcos exteriores con los interiores
  for (m=1; m<=nm; m++) {
    if (enGrupo[m] == 0)
      for (m2=1; m2<=nm; m2++)
        if ((enGrupo[m2] != 0) && enVertical (marcos, m, m2, incl)
                               && disjuntosMar (marcos, m, m2)) {
          enGrupo[m] = enGrupo[m2];
//        printf ("---------1 %3d %3d\n", m, m2);
          }
    // no pudo agrupar
    if (enGrupo[m] == 0)
      for (m2=1; m2<=nm; m2++)
        if ((enGrupo[m2] != 0) && enVertical (marcos, m, m2, incl)) {
          enGrupo[m] = enGrupo[m2];
//        printf ("---------2 %3d %3d\n", m, m2);
          }
    }
  for (m=1; m<=nm; m++) {
    if (enGrupo[m] == 0)
      enGrupo[m] = ++ng ;
    }
  if (d_valori >= 10) {
    printf ("segmenta2.agrupaMarcos enGrupo: ");
    printf ("     alturax: %3d  base: %3d\n", alturax, base);
    for (m=1; m<=nm; m++) printf ("%3d %3d,", m, enGrupo[m]);
    printf ("\n");
    }
  return  ng ;
}

int agrupaMarcos0 (int * enGrupo, tmarco * marcos, int nm, int base, int alturax, int incl)
{
  int m;
  // no quiere agrupar
  // printf ("segmenta2.agrupaMarcos0 nm:%3d\n", nm);
  for (m=1; m<=nm; m++)
    enGrupo[m] = m;
  if (d_valori >= 10)
    printf ("segmenta2.agrupaMarcos0\n");
  return  nm ;
}

void ordenaVentanas (tvvents vvents, int ng) 
{
  int i, j;
  tvent x;
  if (ng > 1) 
    for (i=0; i<(ng-1); i++)
      for (j=i+1; j<ng; j++)
        if (vvents->ventanas[j]->dx < vvents->ventanas[i]->dx){
          x                    = vvents->ventanas [j] ;
          vvents->ventanas [j] = vvents->ventanas [i] ;
          vvents->ventanas [i] = x ;
        }
}

// --------------------  --------------------

void segmentaAfVent (tvvents * nvvents, tvent vent, int umbral, int umbral2)
{
  tvvents vvents;
  int n;
  int nm;
  int ng;
  int ** fondo;
  int nseg;
  tmarco marcos[NSEGMENS];
  int enGrupo [NSEGMENS];
           if (vent->nid == 34846)
              printf ("----------- ventalloc -> nid\n");
  // las coordenadas en  fondo  seran una unidad mayor que en la imagen
  fondo = matalloc (vent->tx + 2, vent->ty + 2);   // para marcarFondo 4 direcc
  marcaFondo (fondo, &nseg, vent, umbral);
  // printf ("segmentaAfVent, tx:%4d ty%4d\n", vent->tx, vent->ty);  //
  // printf ("segmentaAfVent, umbral:%4d umbral2:%4d\n", umbral, umbral2);  //
  if (d_valori >= 14) {
    escribeFondo (fondo, vent->tx, vent->ty);  //
    }
  nm = marcosDeSegmento (marcos, vent, fondo, nseg);
//  escribeFondoYmarco (fondo, vent->tx, vent->ty, marcos[3], vent->incl);  //
  ng = agrupaMarcos0 (enGrupo, marcos, nm, 0, 0, 0) ;
  // creamos las ventanas ... y las imagenes (necesarias)
  vvents = vventsalloc (ng) ;
  for (n=1; n<=ng; n++) {
    vvents->ventanas[n-1] =
          ventanaGrupoSegmentos (vent, fondo, enGrupo, marcos, nm, n, umbral2);
    }
  matfree (fondo, vent->ty + 2);
  liberaMarcosDeSegmento (marcos, nm);
  ordenaVentanas (vvents, ng); 
  * nvvents = vvents;
}

void segmentaAfBarco (tvvents * nvvents, tvent vent)
{
  int umbral2 ;
  int umbral ;
  int porcNeg = 35;   // ult 35  // 65 .. 15
  ponUmbrales (&umbral, &umbral2, vent, porcNeg);
  segmentaAfVent (nvvents, vent, umbral, umbral2);
}

// --------------------  --------------------

void segmentaAfRect (tvvents * nvvents, tvent vent, int base, int alturax, int porcNeg, Boolean agrupa)
{
  int umbral2 ;
  int umbral ;
  tvvents vvents;
//  int porcNeg = 60;     // 20120519  // 52,60,68
//  int porcNeg = 60; 
  int n;
  int nm;
  int ng;
  int ** fondo;
  int nseg;
  tmarco marcos[NSEGMENS];
  int enGrupo [NSEGMENS];
  if (d_valori >= 12) {
    printf ("ocre.segmenta2.segmentaAfRect:\n");
    describeVentana (vent);
    }
  ponUmbrales (&umbral, &umbral2, vent, porcNeg);
  // las coordenadas en  fondo  seran una unidad mayor que en la imagen
  fondo = matalloc (vent->tx + 2, vent->ty + 2);   // para marcarFondo 4 direcc
  // printf ("ocre.segmenta2.segmentaAfRect: sigue 1\n");  // XXx1
  marcaFondo (fondo, &nseg, vent, umbral);
  // printf ("ocre.segmenta2.segmentaAfRect: sigue 2\n");  // XXx1
  // printf ("segmentaAfRect, tx:%4d ty%4d\n", vent->tx, vent->ty);  //
  // escribeFondo (fondo, vent->tx, vent->ty);  //
  nm = marcosDeSegmento (marcos, vent, fondo, nseg);
  // printf ("ocre.segmenta2.segmentaAfRect: sigue 3\n");  // XXx1
//  escribeFondoYmarco (fondo, vent->tx, vent->ty, marcos[3], vent->incl);  //
  if (agrupa)
    ng = agrupaMarcos (enGrupo, marcos, nm, base, alturax, vent->incl) ;
  else
    ng = agrupaMarcos0 (enGrupo, marcos, nm, base, alturax, vent->incl) ;
  // creamos las ventanas ... y las imagenes (necesarias)
  vvents = vventsalloc (ng) ;
  for (n=1; n<=ng; n++) {
    vvents->ventanas[n-1] =
          ventanaGrupoSegmentos (vent, fondo, enGrupo, marcos, nm, n, umbral2);
    }
  matfree (fondo, vent->ty + 2);
  liberaMarcosDeSegmento (marcos, nm);
  ordenaVentanas (vvents, ng); 
  * nvvents = vvents;
}

void segmentaAfIncl (tvvents * nvvents, tvent vent, int base, int alturax, int porcNeg, Boolean agrupa)
{
  int fondo;
  tvent ventl;
  tdcaracHistoGr carHisGr;
  // ventana limpia (e imagen)  ?
  carHisGr = carhisgrDeVentana (vent);
  fondo = carHisGr->xblanco ;
  ventanaLimpConMargenDeVentanaIncl (&ventl, vent, fondo, 5);
  ponerCarhisgrenVentana (ventl, carHisGr);
  ventl->incl = vent->incl ;   // ?
  if (d_valori >= 20) {
    printf ("ocre.segmenta2.segmentaAfIncl: \n");
    describeVentana (ventl);
    }
  //
  segmentaAfRect (nvvents, ventl, base, alturax, porcNeg, agrupa);
}

void segmentaAf (tvvents *nvvents, tglifo glifo, int porcNeg, Boolean agrupa)
{
  tvent vent;
  int base, alturax;
  // int porcNeg = 60; 
  if (d_valori >= 12) {
    printf ("ocre.segmenta2.segmentaAf:\n");
    describeGlifo (glifo, 1);
    }
  vent = glifo->venDeGlifo ;
  // relativas a la imagen base
  baseyalturaxDeGlifo (& base, & alturax, glifo);
  // relativas a la ventana
  alturax -= vent->dy ;
  base    -= vent->dy ;
  if (vent->incl == 0)
    segmentaAfRect (nvvents, vent, base, alturax, porcNeg, agrupa);
  else
    segmentaAfIncl (nvvents, vent, base, alturax, porcNeg, agrupa);
}

// -------------------- x --------------------
