
#include <libintl.h>
#include <locale.h>
#define _(mens) gettext(mens)
#define N_(mens) (mens)

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "base.h"
#include "config.h"
#include "imag.h"
#include "misc2.h"
#include "opcns.h"
#include "tfd2d.h"

/******************    Transformada de Fourier    ******************/

tmatcomplejo nuevaMatrizComplejo (int ty, int tx)
{
  tmatcomplejo m;
  int y;
  m = (tmatcomplejo) malloc (ty * sizeof(tcomplejo *));
  for (y=0; y<ty; y++)
    m[y] = (tcomplejo *) malloc (tx * sizeof(tcomplejo));
  return m;
}

void liberaMatrizComplejo (tmatcomplejo m, int ty)
{
  int y;
  for (y=0; y<ty; y++)
    free (m[y]);
  free (m);
}

tcomplejo v[TF_TAMMAX], vt[TF_TAMMAX];

/****************************** movs ******************************/

static void mueve_fila_imagen_a_vector (b, x0, y0, y, tx)
char **b;
int x0, y0, y, tx;
{
  int x ;
  for (x = 0; x < tx; x++) {
    v[x].pr = (b[y0+y][x+x0] & 0xFF) ;
    v[x].pi = 0.0 ;
/*
    printf ("fiavec: %15.7f %15.7f\n", v[x].pr, v[x].pi) ;
*/
    }
/*
  printf ("\n") ;
*/
}

/* ultimo escrito Thu Dec 13 00:50:08 CET 2001 */
static void mueve_fila_matriz_a_vector (tmatcomplejo m, int y, int tx)
{
  int x ;
  for (x = 0; x < tx; x++) {
    v[x].pr = m [y][x] .pr ;
    v[x].pi = m [y][x] .pi ;
/*
    printf ("matavx: %15.7f %15.7f\n", v[x].pr, v[x].pi) ;
*/
    }
/*
  printf ("\n") ;
*/
}

static void mueve_vector_a_fila_matriz (int y, int tx, tmatcomplejo m)
{
  int x ;
  for (x = 0; x < tx; x++) {
    m [y][x] .pr  = vt [x] .pr ;
    m [y][x] .pi  = vt [x] .pi ;
/*
    printf ("vecmat: %15.7f %15.7f\n", vt[x].pr, vt[x].pi) ;
*/
    }
/*
  printf ("\n") ;
*/
}

static void mueve_columna_matriz_a_vector (tmatcomplejo m, int x, int ty)
{
  int y;
  for (y = 0; y < ty; y++) {
    v[y].pr = m [y][x] .pr ;
    v[y].pi = m [y][x] .pi ;
/*
    printf ("matavy: x:%2d y:%2d %15.7f %15.7f\n", x, y, v[y].pr, v[y].pi) ;
*/
    }
/*
  printf ("\n") ;
*/
}

static void mueve_vector_a_columna_matriz (int x, int ty, tmatcomplejo m)
{
  int y;
  for (y = 0; y < ty; y++) {
    m [y][x] .pr  = vt [y] .pr ;
    m [y][x] .pi  = vt [y] .pi ;
/*
    printf ("vecyamat: %15.7f %15.7f\n", vt[y].pr, vt[y].pi) ;
*/
    }
/*
  printf ("\n") ;
*/
}

/****************************** ..... ******************************/

void imagen_de_matriz (timagf * nimag, tmatcomplejo m, int tx, int ty)
{
  timagf imag;
  int x, y;
  int ng;
  char ** pix;
  creaImagenPlana (& imag, 0, tx, ty);
  pix = imag->ven0->pix ;
  for (y = 0; y < ty; y++)
  for (x = 0; x < tx; x++) {
    ng = m [y][x].pr;
    ng = min (ng, 255);
    ng = max (ng, 0);
    pix [y][x] = ng;
    }
  * nimag = imag;
}

/****************************** tfd_1 ******************************/

double pi=3.14159265358979323846 ;

double sintab[TF_TAMMAX], costab[TF_TAMMAX];

static void atablas (int n)
{
  int i;
  for (i=0; i<=n; i++){
    sintab[i] = sin ((2 * pi * i)/n);
    costab[i] = cos ((2 * pi * i)/n);
    }
}

static void tfd (int t) 
{
  int j, k;
  int jkmodt;
  for (j=0; j<t; j++){
    vt [j].pr = 0.0;
    vt [j].pi = 0.0;
    for (k=0; k<t; k++){
      jkmodt = (j * k) % t;
      vt [j].pr += v [k].pr * costab [jkmodt] -
                   v [k].pi * sintab [jkmodt]   ;
      vt [j].pi += v [k].pi * costab [jkmodt] +
                   v [k].pr * sintab [jkmodt]   ;
      }
    }
}

/* transformada inversa */
/* cambia el signo de la parte 'seno' */
static void tfdi(int t) 
{
  int j, k;
  int jkmodt;
  for (j=0; j<t; j++){
    vt [j].pr = 0.0;
    vt [j].pi = 0.0;
    for (k=0; k<t; k++){
      jkmodt = (j * k) % t;
      vt [j].pr += v [k].pr * costab [jkmodt] +
                   v [k].pi * sintab [jkmodt]   ;
      vt [j].pi += v [k].pi * costab [jkmodt] -
                   v [k].pr * sintab [jkmodt]   ;
      }
    }
}

static void tfd0(t)  /* entrada real */
int t;
{
  int j, k;
  int jkmodt;
  for (j=0; j<t; j++){
    vt [j].pr = 0.0;
    vt [j].pi = 0.0;
    for (k=0; k<t; k++){
      jkmodt = (j * k) % t;
      vt [j].pr += v [k].pr * costab [jkmodt] ;
      vt [j].pi += v [k].pr * sintab [jkmodt] ;
      }
    }
}

/**************************** filtro freqs *************************/

/* tv: tamanio vector, ht: altura triangulo, hd: altura delta */
void preparaVector (double ** nf, int tv, int ht, int hd )
{
  vectint v;
  tveccomplejo vc;
  double * f;
/*
  vector_con_triangulo (&v, tv, ht, ht);
  v[ht] += hd ;
*/
  vector_con_rectangulo (&v, tv, 0, ht, ht);
  v[ht/2] += hd ;
/**/
  transfVector (& vc, v, tv);
  modulosnVeccTr (&f, vc, tv, ht);
  free (v);
  free (vc);
  *nf = f;
}

void filtroFrecuencias (tmatcomplejo m,
                         int tx, int ty)
{
  double * fx, * fy;
  int x, y;
  int ht, hd;
  double k2, k;
  ht = 8 ;
  hd = 128 ;
  preparaVector (&fx, tx, ht, hd);
  preparaVector (&fy, ty, ht, hd);
  k  = 1.0 * (hd + ht * ht) / ( ht * ht) ;
  k2 = k * k ;
  /* filtra */
  for (y=0; y<ty; y++)
  for (x=0; x<tx; x++) {
    m [y][x].pr *= k2 / (fy[y]*fx[x]) ;
    m [y][x].pi *= k2 / (fy[y]*fx[x]) ;
    }
  /* libera vectores */
  free (fx);
  free (fy);
}

/****************************** tfd_2 ******************************/

void transfRectSinNorm (tmatcomplejo *nm, tvent vent,       
                        int x0, int y0, int tx, int ty)
{
  int x, y;
  char ** pix;
  tmatcomplejo m;
  if (tx > TF_TAMMAXX) {
    printf (_("%s: x_dimension (%3d) larger than anticipated (%3d)\n"),
               "ocre.tfd2d.transfRect", tx, TF_TAMMAXX);
    vent->pix[5555][5555]=0;
    exit (9);
    }
  if (ty > TF_TAMMAXY) {
    printf (_("%s: y_dimension (%3d) larger than anticipated (%3d)\n"),
               "ocre.tfd2d.transfRect", ty, TF_TAMMAXY);
    exit (9);
    }
  m = nuevaMatrizComplejo (TF_TAMMAXX, TF_TAMMAXY);
  /* filas */
  pix = vent->pix;
  atablas (tx);
  for (y=0; y<ty; y++){
    mueve_fila_imagen_a_vector (pix, x0, y0, y, tx) ;
    tfd0 (tx) ;
    mueve_vector_a_fila_matriz (y, tx, m);
    }
  /* columnas */
  atablas (ty);
  for (x=0; x<tx; x++){
    mueve_columna_matriz_a_vector (m, x, ty) ;
    tfd (ty) ;
    mueve_vector_a_columna_matriz (x, ty, m) ;
    }
  if (d_valori >= 30)
    printf ("transfRectSinNorm, %d %d %d %d\n", x0, y0, tx, ty) ;
  *nm = m;
}

void transfRectInvSinNorm (tmatcomplejo m,
                           int tx, int ty)
{
  int x, y;
  if (tx > TF_TAMMAXX) {
    printf (_("%s: x_dimension (%3d) larger than anticipated (%3d)\n"),
               "ocre.tfd2d.transfRect", tx, TF_TAMMAXX);
    exit (9);
    }
  if (ty > TF_TAMMAXY) {
    printf (_("%s: y_dimension (%3d) larger than anticipated (%3d)\n"),
               "ocre.tfd2d.transfRect", ty, TF_TAMMAXY);
    exit (9);
    }
  /* filas */
  atablas (tx);
  for (y=0; y<ty; y++){
    mueve_fila_matriz_a_vector (m, y, tx);
    tfdi (tx) ;
    mueve_vector_a_fila_matriz (y, tx, m);
    }
  /* columnas */
  atablas (ty);
  for (x=0; x<tx; x++){
    mueve_columna_matriz_a_vector (m, x, ty) ;
    tfdi (ty) ;
    mueve_vector_a_columna_matriz (x, ty, m) ;
    }
  if (d_valori >= 30)
    printf ("transfRectInvSinNorm, %d %d\n", tx, ty) ;
}

/* normaliza para calcular las caracteristicas */
void transfRectNormCar (tmatcomplejo m, int tx, int ty)
{
  int x, y;
  double  ac, acr2;
  ac = 0.0 ;
  for (x=0; x<tx; x++)
  for (y=0; y<ty; y++)
    if ((x != 0) || (y != 0)){
      ac +=  (m[y][x].pr *  m[y][x].pr) +
             (m[y][x].pi *  m[y][x].pi)   ;
      if (d_valori >= 50)
        printf ("%4d %4d %10.3f\n", x, y, ac) ;
      }
  acr2 = sqrt (ac) / 128.0 ;
  if (d_valori >= 30)
    printf ("%10.3f\n", acr2) ;
  for (x=0; x<tx; x++)
  for (y=0; y<ty; y++){
    m[y][x].pr  /= acr2 ;
    m[y][x].pi  /= acr2 ;
    }
  if (d_valori >= 30)
    printf ("transfRectNormCar, %d %d\n", tx, ty) ;
}

/* normaliza */
void transfRectNorm (tmatcomplejo m, int tx, int ty)
{
  int x, y;
  double  div;
  div = tx*ty ;
  div = sqrt (tx*ty) ;
  if (d_valori >= 30)
    printf ("%10.3f\n", div) ;
  for (x=0; x<tx; x++)
  for (y=0; y<ty; y++){
    m[y][x].pr  /= div ;
    m[y][x].pi  /= div ;
    }
  if (d_valori >= 30)
    printf ("transfRectNorm, %d %d\n", tx, ty) ;
}

void transfRect (tmatcomplejo *nm, tvent vent,       
                 int x0, int y0, int tx, int ty)
{
  tmatcomplejo m;
  transfRectSinNorm ( &m, vent, x0, y0, tx, ty);
  transfRectNormCar (  m,               tx, ty);
  if (d_valori >= 30)
    printf ("transfRect, %d %d %d %d\n", x0, y0, tx, ty) ;
  *nm = m;
}

/* .. Directa, Filtro_frecuencias, Inversa */
void transfRectDFI (tmatcomplejo *nm, tvent vent,       
                    int x0, int y0, int tx, int ty)
{
  tmatcomplejo m;
  transfRectSinNorm    ( &m, vent, x0, y0, tx, ty);
  transfRectNorm       (  m,               tx, ty);
  /*
  escribe1tr (m, tx, ty);
  */
  filtroFrecuencias    (  m,               tx, ty);
  transfRectInvSinNorm (  m,               tx, ty);
  transfRectNorm       (  m,               tx, ty);
  if (d_valori >= 30)
    printf ("transfRectDI, %d %d %d %d\n", x0, y0, tx, ty) ;
  *nm = m;
}

void transfRectDI (tmatcomplejo *nm, tvent vent,       
                   int x0, int y0, int tx, int ty)
{
  tmatcomplejo m;
  transfRectSinNorm    ( &m, vent, x0, y0, tx, ty);
  transfRectNorm       (  m,               tx, ty);
  /*
  escribe1tr (m, tx, ty);
  */
  transfRectInvSinNorm (  m,               tx, ty);
  transfRectNorm       (  m,               tx, ty);
  if (d_valori >= 30)
    printf ("transfRectDI, %d %d %d %d\n", x0, y0, tx, ty) ;
  *nm = m;
}

void transfVentDI (timagf *nimag, tvent vent)
{
  tmatcomplejo m;
  timagf imag;
  transfRectDFI (&m, vent, 0, 0, vent->tx, vent->ty);
  /*
  escribe1tr (m, vent->tx, vent->ty);
  */
  imagen_de_matriz (&imag, m, vent->tx, vent->ty);
  * nimag = imag;
}

void transfVent (tmatcomplejo *nm, tvent vent)
{
  transfRect (nm, vent, 0, 0, vent->tx, vent->ty);
}

void transfVent0 (tmatcomplejo *nm, tvent vent)
{
  int tx, ty;
  timagf imaga, imagr;
  tx = vent->tx;
  ty = vent->ty;
  if ((tx > TF_TAMMAXY) || (ty > TF_TAMMAXY)) {
    /* vent  muy grande */
    /*
    creaImagenReducida (&imagr, vent->imaBase);
					*/
    creaImagenReducidaDeVentana (&imagr, vent, (tx-1)/TF_TAMMAXY + 1,
		                               (ty-1)/TF_TAMMAXY + 1);
    transfVent (nm, imagr->ven0);
    destruyeImagen (imagr);
    }
  else {
    // if (tx >= 6 && ty >= 6)    /* 6  viene de tomar hasta 5 comp en tf */
    if (tx >= 7 && ty >= 7)    // 7  viene de tomar hasta 6 comp en tf
      transfVent (nm, vent);
    else {
      // creaImagenAmpliada (&imaga, vent, (6-1)/tx + 1, (6-1)/ty + 1);
      creaImagenAmpliada (&imaga, vent, (7-1)/tx + 1, (7-1)/ty + 1);
      transfVent (nm, imaga->ven0);
      destruyeImagen (imaga);
      }
    }
}

void escribe1tr (tmatcomplejo m, int nx, int ny)
{
  int x, y;
  for (y=0; y<ny; y++){
    for (x=0; x<nx; x++)
      printf ("%7.1f ", (m[y][x].pr));
    printf ("\n") ;
    for (x=0; x<nx; x++)
      printf ("%7.1f ", (m[y][x].pi));
    printf ("\n") ;
    printf ("\n") ;
    }
  printf ("\n") ;
}

/****************************** ..... ******************************/

tveccomplejo nuevoVectorComplejo (int t)
{
  return (tveccomplejo) malloc (t * sizeof(tcomplejo));
}

void liberaVectorComplejo (tveccomplejo v)
{
  free (v);
}

static void mueve_entrada_real_a_vector (int *vec, int t)
{
  int i ;
  for (i = 0; i < t; i++) {
    v[i].pr = vec[i] ;
    v[i].pi = 0.0 ;
    }
}

static void mueve_vector_tfd_a_salida (tcomplejo *vs, int t)
{
  int i ;
  for (i = 0; i < t; i++) {
    vs[i].pr = vt[i].pr ;
    vs[i].pi = vt[i].pi ;
/*
    printf ("vector_tfd_a_sal: %4d %15.7f %15.7f\n", i, vs[i].pr, vs[i].pi) ;
*/
    }
}

/* normaliza para calcular las caracteristicas */
void transfVectNormSeg (tveccomplejo vc, int tx)
{
  int x;
  double  ac, acr2;
  ac = 0.0 ;
  for (x=1; x<tx; x++){
    ac +=  (vc[x].pr *  vc[x].pr) +
           (vc[x].pi *  vc[x].pi)   ;
    if (d_valori >= 50)
      printf ("%4d %10.3f\n", x, ac) ;
      }
  acr2 = sqrt (ac) / 128.0 ;
  if (d_valori >= 30)
    printf ("%10.3f\n", acr2) ;
  for (x=0; x<tx; x++){
    vc[x].pr  /= acr2 ;
    vc[x].pi  /= acr2 ;
    }
  if (d_valori >= 30)
    printf ("transfVectNormSeg, %d\n", tx) ;
}

void transfVector (tveccomplejo *nv, int *vec, int t)
{
  tveccomplejo vs;
  if (t > TF_TAMMAX) {
    printf (_("%s: dimension (%4d) larger than anticipated (%4d)\n"),
               "ocre.tfd2d.transfVector", t, TF_TAMMAX);
    exit (9);
    }
  vs = nuevoVectorComplejo (t);
  mueve_entrada_real_a_vector (vec, t);
  atablas (t);
  tfd0 (t);
  mueve_vector_tfd_a_salida (vs, t);
  *nv = vs;
  if (d_valori >= 10)
    printf ("ocre: tfd2d.transfVector\n") ;
}

/* solo de la mitad, supongo que parto de la tfd de reales */
/* y normalizado para que la ~energia (mitad de indices, sin 0) sea 255^2 */
void modulosnVecc (int **nvm, tveccomplejo vc, int t)
{
  int *mods;
  int i;
  double sumcuad, norm;
  sumcuad = 0.0;
  for(i=1;i<t/2;i++)
    sumcuad += vc[i].pr*vc[i].pr + vc[i].pi*vc[i].pi ;
  norm = 255.0 / sqrt(sumcuad);
  mods = (int *) malloc (((t/2) + 1) * sizeof (int));
  for(i=1;i<t/2;i++) {
    mods[i] = sqrt(vc[i].pr*vc[i].pr +
                   vc[i].pi*vc[i].pi  ) * norm + 0.5 ;
    }
  *nvm = mods ;
  if (d_valori >= 10)
    printf ("ocre: tfd2d.modulosnVecc\n") ;
}

void modulosnVeccTr (double **nvm, tveccomplejo vc, int t, int alt)
{
  double *mods;
  int i;
  double norm;
  norm = 1.0 / (alt * alt) ;
  mods = (double *) malloc (t * sizeof (double));
  for(i=0;i<t;i++) {
    mods[i] = sqrt(vc[i].pr*vc[i].pr +
                   vc[i].pi*vc[i].pi  ) * norm ;
    }
  *nvm = mods ;
  if (d_valori >= 10)
    printf ("ocre: tfd2d.modulosnVeccTr\n") ;
}

void escribe0tr (tveccomplejo v, int t)
{
  int x;
  for (x=0; x<t; x++)
    printf ("%7.1f ", (v[x].pr));
  printf ("\n") ;
  for (x=0; x<t; x++)
    printf ("%7.1f ", (v[x].pi));
  printf ("\n") ;
  printf ("\n") ;
}
